from ..runtime import parameter
import numpy as np
import tempfile


def test_override_new():
    with tempfile.NamedTemporaryFile('w') as values:
        values.write(
            "[parameters]\n"
            "p1=-3.0  0.0  3.0\n"
            "p2=-3.0  0.0  3.0\n"
            "p3=20.0\n")
        values.flush()
        print(values.name)
        params = parameter.Parameter.load_parameters(values.name, override=None)

    assert params[0] == ("parameters", "p1")
    assert params[0].start == 0.0
    assert np.allclose(params[0].limits, [-3, 3])

    assert params[1] == ("parameters", "p2")
    assert params[1].start == 0.0

    assert params[2] == ("parameters", "p3")
    assert np.isclose(params[2].start, 20.0)


    with tempfile.NamedTemporaryFile('w') as values:
        values.write(
            "[parameters]\n"
            "p1=-3.0  0.0  3.0\n"
            "p2=-3.0  0.0  3.0\n"
            "p3=20.0\n")
        values.flush()
        override = {
            ("parameters", "p1"): "1.0",
            ("xxx", "aaa"): "100.0",
        }
        params = parameter.Parameter.load_parameters(values.name, override=override)
    

    assert params[0] == ("parameters", "p1")
    assert params[0].start == 1.0
    assert np.allclose(params[0].limits, 1.0)

    assert params[1] == ("parameters", "p2")
    assert params[1].start == 0.0
    assert np.allclose(params[1].limits, [-3, 3])

    assert params[2] == ("parameters", "p3")
    assert np.isclose(params[2].start, 20.0)

    print(params)
    assert params[3] == ("xxx", "aaa")
    assert np.isclose(params[3].start, 100.0)
    assert np.allclose(params[3].limits, 100.0)




from .. import ParallelSampler
import numpy as np
import sys
import os

def log_probability_function(p):
    r = pipeline.run_results(p)
    out = [r.like, r.prior]

    # Flatten any vector outputs here
    for e in r.extra:
        if np.isscalar(e):
            out.append(e)
        else:
            out.extend(e)
    out = tuple(out)
    return out


def prior_transform(p):
    return pipeline.denormalize_vector_from_prior(p)


class NautilusSampler(ParallelSampler):
    parallel_output = False
    internal_resume = True
    sampler_outputs = [ ('log_weight', float), ('prior', float), ("post", float)]

    def config(self):
        global pipeline
        pipeline = self.pipeline

        if self.is_master():
            self.n_live = self.read_ini("n_live", int, 1500)
            self.n_update = self.read_ini("n_update", int, self.n_live)
            self.enlarge = self.read_ini(
                "enlarge", float, 1.1**self.pipeline.nvaried)
            self.n_batch = self.read_ini("n_batch", int, 100)
            self.random_state = self.read_ini("random_state", int, -1)
            if self.random_state < 0:
                self.random_state = None
            self.resume_ = self.read_ini("resume", bool, False)
            self.f_live = self.read_ini("f_live", float, 0.01)
            self.n_shell = self.read_ini("n_shell", int, self.n_batch)
            self.n_eff = self.read_ini("n_eff", float, 10000.0)
            self.discard_exploration = self.read_ini(
                "discard_exploration", bool, False)
            self.verbose = self.read_ini("verbose", bool, False)

        self.converged = False

    def execute(self):
        from nautilus import Sampler

        n_dim = self.pipeline.nvaried

        try:
            resume_filepath = self.output.name_for_sampler_resume_info()
        except NotImplementedError:
            resume_filepath = None

        if resume_filepath is not None:
            resume_filepath = resume_filepath + ".hdf5"
            if self.resume_ and os.path.exists(resume_filepath):
                print(f"Resuming Nautilus from file {resume_filepath}")

        sampler = Sampler(
            prior_transform,
            log_probability_function,
            n_dim,
            n_live=self.n_live,
            n_update=self.n_update,
            enlarge=self.enlarge,
            n_batch=self.n_batch,
            random_state=self.random_state,
            filepath=resume_filepath,
            resume=self.resume_,
            pool=self.pool,
            blobs_dtype=float
        )

        sampler.run(f_live=self.f_live,
                    n_shell=self.n_shell,
                    n_eff=self.n_eff,
                    discard_exploration=self.discard_exploration,
                    verbose=self.verbose)

        for sample, logwt, logl, blob in zip(*sampler.posterior(return_blobs=True)):
            blob = np.atleast_1d(blob)
            prior = blob[0]
            extra = blob[1:]
            logp = logl + prior
            self.output.parameters(sample, extra, logwt, prior, logp)

        self.output.final(
            "efficiency", sampler.effective_sample_size() / sampler.n_like)
        self.output.final("neff", sampler.effective_sample_size())
        self.output.final("nsample", len(sampler.posterior()[0]))
        self.output.final("log_z", sampler.evidence())
        self.converged = True

    def is_converged(self):
        return self.converged



from .. import ParallelSampler
import numpy as np


def log_likelihood(p):
    like, _ = pipeline.likelihood(p)
    # PocoMC cannot cope with infinities
    if not np.isfinite(like):
        like = -1e30
    return like


def log_prior(p):
    return pipeline.prior(p)


class PocoSampler(ParallelSampler):
    parallel_output = False
    supports_resume = False
    sampler_outputs = [("prior", float), ("post", float)]

    def config(self):
        global pipeline
        pipeline = self.pipeline

        if self.is_master():
            import pocomc

            self.pocomc = pocomc

            # Parameters of the sampler
            self.ndim = len(self.pipeline.varied_params)
            self.nparticles = self.read_ini("n_particles", int)
            self.nsample_add = self.read_ini("add_samples", int, 0)
            self.threshold = self.read_ini("threshold", float, 1.0)
            self.scale = self.read_ini("scale", bool, True)
            self.rescale = self.read_ini("rescale", bool, False)
            self.diagonal = self.read_ini("diagonal", bool, True)
            self.ess = self.read_ini("ess", float, 0.95)
            self.gamma = self.read_ini("gamma", float, 0.75)
            self.n_min = self.read_ini("n_min", int, self.ndim // 2)
            self.n_max = self.read_ini("n_max", int, self.ndim * 10)
            seed = self.read_ini("seed", int, 0)
            if seed == 0:
                seed = None

            # Starting positions and values for the chain
            self.num_samples = 0
            self.prob0 = None

            self.p0 = np.array(
                [self.pipeline.randomized_start() for i in range(self.nparticles)]
            )
            self.output.log_info(
                "Generating random starting positions from within prior"
            )

            bounds = np.array([p.limits for p in self.pipeline.varied_params])

            # Finally we can create the sampler
            self.sampler = self.pocomc.Sampler(
                n_dim=self.ndim,
                n_particles=self.nparticles,
                log_likelihood=log_likelihood,
                log_prior=log_prior,
                random_state=seed,
                threshold=self.threshold,
                scale=self.scale,
                rescale=self.rescale,
                diagonal=self.diagonal,
                vectorize_likelihood=False,
                vectorize_prior=False,
                infer_vectorization=False,
                pool=self.pool,
                bounds=bounds,
            )

            self.converged = False

            for sec, name in pipeline.extra_saves:
                col = "{}--{}".format(sec, name)
                print(
                    "WARNING: POCO DOES NOT SUPPORT DERIVED PARAMS - NOT SAVING {}".format(
                        col
                    )
                )
                self.output.del_column(col)

    def execute(self):
        # There's no checkpointing yet
        self.sampler.run(
            self.p0, n_min=self.n_min, n_max=self.n_max, gamma=self.gamma, ess=self.ess
        )

        if self.nsample_add > 0:
            self.sampler.add_samples(self.nsample_add)

        results = self.sampler.results

        samples = results["samples"]
        likes = results["loglikelihood"]
        priors = results["logprior"]
        n = len(samples)
        outputs = []
        for (p, prior, like) in zip(samples, priors, likes):
            post = prior + like
            self.output.parameters(p, prior, post)

        self.output.final("log_z", results["logz"][-1])
        self.converged = True

    def is_converged(self):
        return self.converged
